/** Three best exemplar helper
 * A best exemplar finder.  Scans over the entire image (using a
 * sliding window) and finds the exemplar which minimizes the sum
 * squared error (SSE) over the to-be-filled pixels in the target
 * patch. 
 *
 */
#include "mex.h"
#include <limits.h>

// This is the main computational subroutine
void bestexemplarhelper(const int mm, const int nn, const int m, const int n, 
			const double *img, const double *Ip, 
			const mxLogical *toFill, const mxLogical *sourceRegion,
			double *best) 
{
  register int i,j,ii,jj,ii2,jj2,M,N,I,J,ndx,ndx2,ndx_d,ndx_d2,mn=m*n,mmnn=mm*nn;
  double patchErr=0.0,err=0.0,err_d=0.0,bestErr=1000000000.0,bestErr2=1000000001.0,bestErr3=1000000002.0,bestErr4=1000000003.0,bestErr5=1000000004.0;

  /* foreach patch */
  N=nn-n+1;  M=mm-m+1; // so that patch remains within the image: nn,mm=image; n,m=patch.
  for (j=1; j<=N; ++j) {
    J=j+n-1;
    for (i=1; i<=M; ++i) {
      I=i+m-1;
  /*** Calculate patch error ***/
  /* foreach pixel in the current patch */
  for (jj=j,jj2=1; jj<=J; ++jj,++jj2) { // jj,ii corresponds to current pixel location
	for (ii=i,ii2=1; ii<=I; ++ii,++ii2) {
	  ndx=ii-1+mm*(jj-1); // ndx corresponds to current pixel in the image
	  ndx_d=ndx;
	  ndx_d2=ndx+1;
	  if (!sourceRegion[ndx]) // if it doesn't lie in sourceRegion or if it lies in fillRegion
	    goto skipPatch;
	  ndx2=ii2-1+m*(jj2-1);   // ndx2 corresponds to image mask 9*9
	  if (!toFill[ndx2]) {    // if ndx2 doesn't belong to tofill of the target patch
		 if (jj<J && ii<I)
		  {
			  //printf("diff pixels %d \t %d \n",ndx_d,ndx_d2);
			  err_d=(img[ndx_d]+img[ndx_d+=mmnn]+img[ndx_d+=mmnn])/3-(img[ndx_d2]+img[ndx_d2+=mmnn]+img[ndx_d2+=mmnn])/3;
			  if (err_d < 127)
			  {err_d = 0;}
			  else 
			  {err_d = 255;}
		 }
		//printf("patch pixels %d \n",ndx);
		patchErr+= err*err;
		//printf("diff cost is %f \n",patchErr);
		err=img[ndx      ] - Ip[ndx2    ]; patchErr += err*err;
	    err=img[ndx+=mmnn] - Ip[ndx2+=mn]; patchErr += err*err;
	    err=img[ndx+=mmnn] - Ip[ndx2+=mn]; patchErr += err*err;
		//printf("pixel value cost is %f \n",patchErr);
	  }}}

  /*** Update ***/
  if (patchErr < bestErr) {
	 bestErr = patchErr; 
	 best[20] = bestErr;
	 best[0] = i; best[1] = I;
	 best[2] = j; best[3] = J;}
	 else if (patchErr < bestErr2) {
			best[21] = bestErr;
			bestErr2 = patchErr;
			best[4] = i; best[5] = I;
			best[6] = j; best[7] = J;}
	 else if (patchErr < bestErr3) {
			best[22] = bestErr;
			bestErr3 = patchErr;
			best[8] = i; best[9] = I;
			best[10] = j; best[11] = J;}
	 else if (patchErr < bestErr4) {
			best[23] = bestErr;
			bestErr4 = patchErr;
			best[12] = i; best[13] = I;
			best[14] = j; best[15] = J;}
	 else if (patchErr < bestErr5) {
			best[24] = bestErr;
			bestErr5 = patchErr;
			best[16] = i; best[17] = I;
			best[18] = j; best[19] = J;}
  
  /*** Reset ***/
  skipPatch:
      patchErr = 0.0;
    }}
}


// This subroutine is similar to the main function of the C-Program
/* best = bestexemplarhelper(mm,nn,m,n,img,Ip,toFill,sourceRegion); */
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[]) 
{
  int mm,nn,m,n;
  double *img,*Ip,*best;
  mxLogical *toFill,*sourceRegion;  //All logical mxArrays store their data elements as mxLogical rather than as bool.

  /* Extract the inputs *///best = bestexemplarhelper(mm,nn,m,n,img,img(rows,cols,:),toFill',sourceRegion);
  mm = (int)mxGetScalar(prhs[0]);// mm,nn are the size of the original image
  nn = (int)mxGetScalar(prhs[1]);//m,n are the size of the small patch image which is to be inpainted
  m  = (int)mxGetScalar(prhs[2]);//Ip is the small patch of the image which is to be inpainted
  n  = (int)mxGetScalar(prhs[3]);
  img = mxGetPr(prhs[4]);
  Ip  = mxGetPr(prhs[5]);
  toFill = mxGetLogicals(prhs[6]);
  sourceRegion = mxGetLogicals(prhs[7]);
  
  /* Setup the output */
  plhs[0] = mxCreateDoubleMatrix(25,1,mxREAL);
  best = mxGetPr(plhs[0]);
  best[0]=best[1]=best[2]=best[3]=best[4]=best[5]=best[6]=best[7]=best[8]=best[9]=best[10]=best[11]=best[12]=best[13]=best[14]=0.0;
  best[15]=best[16]=best[17]=best[18]=best[19]=best[20]=best[21]=best[22]=best[23]=best[24]=0.0;
  /* Do the actual work */
  bestexemplarhelper(mm,nn,m,n,img,Ip,toFill,sourceRegion,best);
}
