
%% This is the code for Image Inpainting through Metric Labelling Via Guided Patch Mixing
%% Refer to and cite the following paper:
% Kumar, Veepin, Jayanta Mukherjee, and Shyamal Kumar Das Mandal.
% "Image Inpainting Through Metric Labeling via Guided Patch Mixing."
% IEEE Transactions on Image Processing 25.11 (2016): 5212-5226.

%% Input parameters
clc;
clear all;
close all;
close all hidden;
I_S=150;  %Image Size
w=4;      %patch size=2*w+1;

%% My parameters
step=3;
lambda=10;   % gives weightage between NC and TVC
lambda2=1000;  % gives weightage between NC and Edge_cost
eta =765;
%eta=22500;  % threshold for a pixel to be an edge pixel in TV_COST
C=1;         % parameter of simulated annealing

%%
dbstop if error

[filename, pathname] = uigetfile;
img = imread(sprintf('%s%s',pathname,filename)); img=imresize(img,[I_S I_S]);
xum=figure;imshow(uint8(img),[]);saveas(xum,'original','jpg');

img_orig =img;

%% for object and scratch removal
close all hidden;
img_orig2 = img;
disp('Mark Target Region, Press mouse button to mark more regions and press any key to finish marking');
[TrBi,~] = MarkRegion(img);

TrBi1 = TrBi;
TrBi = TrBi .*255;
fillRegion1 = TrBi;
fillImg1 = img .* TrBi;
fillImg1 = fillImg1 +img;

figure; imshow(fillImg1);
figure; imshow(fillRegion1);

fillImg = fillImg1;
fillRegion = fillRegion1;

TrBi(:,:,1) = 0;
ATr = ~ TrBi;
ATr = uint8(ATr);
imgTr = img .* ATr;
%imgTr = imgTr + TrBi;
figure; imshow(imgTr);

%% Get one pixel thick target region
fillRegion=rgb2gray(fillRegion);
img=fillImg; %Let us take target image as input
xum=figure;imshow(uint8(img),[]);saveas(xum,'target','jpg');
figure;imshow(fillRegion);title('Region to be filled');


%% Mark the source region
sourceRegion = ~fillRegion;
fillRegion = ~sourceRegion;  % this makes fillRegion logical(only 0 and 1)
binaryImage3D(:,:,2) = uint8(sourceRegion);  binaryImage3D(:,:,1) = uint8(sourceRegion);  binaryImage3D(:,:,3) = uint8(sourceRegion); 
binaryImage3D1 = binaryImage3D;
binaryImage3D(:,:,1) = 0; binaryImage3D(:,:,3) = 0;
SrIm = img_orig .*binaryImage3D + img_orig .* (uint8(~binaryImage3D1));

fillRegion_o=fillRegion;
fillRegion1=fillRegion;
img = double(img);

img_gen=img;

%% Define ind
s=size(img);sz = [size(img,1) size(img,2)]; 
ind=reshape(1:s(1)*s(2),s(1),s(2));
ind2=ind;  ind3=ind;
ind2_test=ind;

%% Seed 'rand' for reproducible results (good for testing)
rand('state',0);
fprintf('preprocessing time is:');

%% Get Five inpainted images using different priority measures
[img1,img2,img3,img4,img5] = genetic_criminisi_2(fillRegion,img,ind,ind2,ind2_test,w,I_S,sz,5);

%% Divide image into two image grids and get the target patch locations
k=1;
for i=w+1:step:I_S-(w)
    for j=w+1:step:I_S-(w)
        if fillRegion_o(i-w:i+w,j-w:j+w)==0
        else  % if it belongs to fillRegion
           fillRegion1(i-w:i+w,j-w:j+w)=200;
           tp(k)= (j-1)*I_S+i; %tp will contain the central pixel of all the target patches so obtained
           k=k+1;
         end
     end
end

fig_no=0;

%% Assign label 1 to each target patch
img_gen=img1;
[m,n]=size(tp);
tp_l(1:m,1:n)=1;

%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Simulated annealing
%tic;
image_count=0;quality_mat=0;cost_mat=0; 
good_image_count=1;good_quality_mat=0;good_cost_mat=0;

best_img=img_gen;

%% Do 2 label assignment according to simulated annealing, same as above
[img_gen,quality_mat,cost_mat,good_cost_mat,image_count,best_img,good_image_count,good_quality_mat]=sa_sub(n,tp,I_S,step,w,fillRegion,img1,img2,img_gen,C,lambda,lambda2,tp_l,quality_mat,cost_mat,good_cost_mat,image_count,good_image_count,good_quality_mat,best_img,2);
[img_gen,quality_mat,cost_mat,good_cost_mat,image_count,best_img,good_image_count,good_quality_mat]=sa_sub(n,tp,I_S,step,w,fillRegion,img_gen,img3,img_gen,C,lambda,lambda2,tp_l,quality_mat,cost_mat,good_cost_mat,image_count,good_image_count,good_quality_mat,best_img,3);
[img_gen,quality_mat,cost_mat,good_cost_mat,image_count,best_img,good_image_count,good_quality_mat]=sa_sub(n,tp,I_S,step,w,fillRegion,img_gen,img4,img_gen,C,lambda,lambda2,tp_l,quality_mat,cost_mat,good_cost_mat,image_count,good_image_count,good_quality_mat,best_img,4);
[img_gen,quality_mat,cost_mat,good_cost_mat,image_count,best_img,good_image_count,good_quality_mat]=sa_sub(n,tp,I_S,step,w,fillRegion,img_gen,img5,img_gen,C,lambda,lambda2,tp_l,quality_mat,cost_mat,good_cost_mat,image_count,good_image_count,good_quality_mat,best_img,5);

xum=figure;imshow(uint8(best_img),[]);saveas(xum,'best_img','jpg');
xum=figure;plot(1:good_image_count-1,good_quality_mat(1:good_image_count-1));title('BRISQUEScorewithIterations');saveas(xum,'quality_count_b','jpg');
